from typing import Optional, Protocol


class RouteProto(Protocol):
    bucket: str
    method: str
    url: str

    def __init__(
        self,
        method: str,
        path: str,
        guild_id: Optional[int] = None,
        channel_id: Optional[int] = None,
        webhook_id: Optional[int] = None,
        webhook_token: Optional[str] = None,
    ) -> None:
        ...



from typing import Literal, TypeVar, Union


class Unset:
    def __bool__(self) -> Literal[False]:
        return False


UNSET = Unset()

T = TypeVar("T", covariant=True)

UnsetOr = Union[T, Unset]



from aiohttp import ClientResponse


class OrxError(Exception):
    """
    Base class for all Orx errors.
    """

    pass


class HTTPError(OrxError):
    """
    Base class for all HTTP errors.
    """

    def __init__(self, response: ClientResponse, *args: object) -> None:
        self.response = response
        self.status = response.status

        super().__init__(*args)


class BadRequest(HTTPError):
    pass


class Unauthorized(BadRequest):
    pass


class Forbidden(BadRequest):
    pass


class NotFound(BadRequest):
    pass


class MethodNotAllowed(BadRequest):
    pass


class UnprocessableEntity(BadRequest):
    pass


class TooManyRequests(BadRequest):
    pass


class ServerError(HTTPError):
    pass


class BadGateway(ServerError):
    pass


class ServiceUnavailable(ServerError):
    pass


class GatewayTimeout(ServerError):
    pass


class GatewayReconnect(OrxError):
    """
    Raised when a shard needs to reconnect to the gateway.
    """

    pass


class GatewayCriticalError(OrxError):
    """
    Raised when a shard encounters a critical error.
    """

    def __init__(self, code: int) -> None:
        self.code = code

        super().__init__(f"Gateway disconnected with close code {code}")



from .client import HTTPClientProto
from .ratelimiter import BucketProto, RatelimiterProto
from .route import RouteProto

__all__ = (
    "BucketProto",
    "HTTPClientProto",
    "RatelimiterProto",
    "RouteProto",
)



from asyncio import Event, Lock, create_task, sleep
from typing import Any

from orx.proto.http import BucketProto


class Bucket:
    __slots__ = ("_rate", "_per", "_lock", "_deferred")

    def __init__(self, rate: int, per: int) -> None:
        self._rate = rate
        self._per = per

        self._deferred = False

        self._lock = Lock()

    async def __aenter__(self) -> "Bucket":
        await self._lock.acquire()
        return self

    async def __aexit__(self, *exc: Any) -> None:
        if not self._deferred:
            self._lock.release()

    async def _unlock(self, after: float) -> None:
        await sleep(after)

        self._lock.release()
        self._deferred = False

    async def set_rate(self, rate: int, per: int) -> None:
        self._rate = rate
        self._per = per

    async def defer(self, unlock_after: float) -> None:
        self._deferred = True
        create_task(self._unlock(unlock_after))


class Ratelimiter:
    __slots__ = (
        "_buckets",
        "_global",
    )

    def __init__(self) -> None:
        self._buckets: dict[str, BucketProto] = {}

        self._global = Event()
        self._global.set()

    async def acquire(self, bucket: str) -> BucketProto:
        if bucket not in self._buckets:
            self._buckets[bucket] = Bucket(rate=1, per=1)
        return self._buckets[bucket]

    async def _unlock_global(self, after: float) -> None:
        await sleep(after)
        self._global.set()

    async def set_global_lock(self, unlock_after: float) -> None:
        if not self._global.is_set():
            raise RuntimeError("Global lock is already locked.")

        self._global.clear()
        create_task(self._unlock_global(unlock_after))



from asyncio import Semaphore, create_task, sleep


class GatewayRatelimiter:
    def __init__(self, rate: int, per: int) -> None:
        self._per = per
        self._lock = Semaphore(rate)

    async def _release(self, after: float) -> None:
        await sleep(after)
        self._lock.release()

    async def acquire(self) -> None:
        await self._lock.acquire()

        create_task(self._release(self._per))



from dataclasses import dataclass
from typing import Any, Optional

from discord_typings.gateway import GatewayEvent as GatewayEventType

from orx.proto.gateway import ShardProto


@dataclass(frozen=True, slots=True)
class GatewayEvent:
    shard: ShardProto
    op: int
    data: Optional[GatewayEventType] = None
    s: Optional[int] = None
    t: Optional[str] = None

    @property
    def d(self) -> Any:
        return self.data["d"]  # type: ignore

    @property
    def dispatch_name(self) -> str:
        if self.t:
            return self.t.upper()
        return f"OP_{self.op}"



from typing import Any, Protocol


class BucketProto(Protocol):
    def __init__(self, rate: int, per: int) -> None:
        ...

    async def __aenter__(self) -> "BucketProto":
        ...

    async def __aexit__(self, *exc: Any) -> None:
        ...

    async def set_rate(self, rate: int, per: int) -> None:
        ...

    async def defer(self, unlock_after: float) -> None:
        ...


class RatelimiterProto(Protocol):
    async def acquire(self, bucket: str) -> BucketProto:
        ...

    async def set_global_lock(self, unlock_after: float) -> None:
        ...



from asyncio import Event
from typing import Any, Callable, Coroutine, Protocol, Type, Union

from discord_typings.gateway import (
    HeartbeatCommand,
    IdentifyCommand,
    RequestGuildMembersCommand,
    ResumeCommand,
    UpdatePresenceCommand,
    VoiceUpdateCommand,
)

from ..http import HTTPClientProto
from .ratelimiter import GatewayRatelimiterProto

Command = Union[
    ResumeCommand,
    IdentifyCommand,
    HeartbeatCommand,
    VoiceUpdateCommand,
    UpdatePresenceCommand,
    RequestGuildMembersCommand,
]


class ShardProto(Protocol):
    id: int
    latency: float | None
    callbacks: dict[str, list[Callable[..., Coroutine[Any, Any, None]]]]
    ready: Event

    def __init__(
        self,
        id: int,
        shard_count: int,
        token: str,
        intents: int,
        ratelimiter_cls: Type[GatewayRatelimiterProto],
    ) -> None:
        ...

    async def connect(self, url: str, http: HTTPClientProto) -> None:
        ...

    async def close(self) -> None:
        ...

    async def send(self, data: Command) -> None:
        ...



from typing import TYPE_CHECKING, Any, Optional, Protocol

from aiohttp import ClientResponse, ClientWebSocketResponse

from orx.impl.types import UNSET, UnsetOr

from .ratelimiter import RatelimiterProto
from .route import RouteProto

if TYPE_CHECKING:
    from orx.impl.http.file import File
else:
    File = Any


class HTTPClientProto(Protocol):
    def __init__(
        self,
        token: Optional[str] = None,
        *,
        api_url: Optional[str] = None,
        default_headers: Optional[dict[str, str]] = None,
        max_retries: Optional[int] = None,
        ratelimiter: Optional[RatelimiterProto] = None,
    ) -> None:
        ...

    async def spawn_websocket(self, url: str, **kwargs: Any) -> ClientWebSocketResponse:
        ...

    async def request(
        self,
        route: RouteProto,
        query_params: Optional[dict[str, str | int]] = None,
        headers: Optional[dict[str, str]] = None,
        *,
        max_retries: Optional[int] = None,
        files: UnsetOr[list[File]] = UNSET,
        json: UnsetOr[Any] = UNSET,
        reason: Optional[str] = None,
    ) -> ClientResponse:
        ...



from typing import Optional


class Route:
    """
    Represents a route to a Discord API route/path, with an HTTP verb.
    """

    __slots__ = (
        "method",
        "url",
        "bucket",
    )

    def __init__(
        self,
        method: str,
        path: str,
        guild_id: Optional[int] = None,
        channel_id: Optional[int] = None,
        webhook_id: Optional[int] = None,
        webhook_token: Optional[str] = None,
    ) -> None:
        self.method = method
        self.url = path.format(
            guild_id=guild_id,
            channel_id=channel_id,
            webhook_id=webhook_id,
            webhook_token=webhook_token,
        )

        webhook_bucket: Optional[str] = None
        if webhook_id:
            webhook_bucket = f"{webhook_id}:{webhook_token}"

        self.bucket = f"{path}-{guild_id}:{channel_id}:{webhook_bucket}"



from asyncio import Future, Task, create_task
from typing import Any, Callable, Coroutine, Optional, Type, TypedDict

from orx.impl.http import Route
from orx.proto.gateway import GatewayRatelimiterProto, ShardProto
from orx.proto.http import HTTPClientProto

from .ratelimiter import GatewayRatelimiter
from .shard import Shard


class SessionStartLimit(TypedDict):
    total: int
    remaining: int
    reset_after: int
    max_concurrency: int


class GetGatewayBot(TypedDict):
    url: str
    shards: int
    session_start_limit: SessionStartLimit


class GatewayClient:
    __slots__ = (
        "shards",
        "_shard_tasks",
        "_shard_ids",
        "_shard_count",
        "_hooks",
        "_token",
        "_intents",
        "_http",
        "_ratelimiter_cls",
        "_shard_cls",
        "_done",
    )

    def __init__(
        self,
        token: str,
        intents: int,
        http: HTTPClientProto,
        ratelimiter_cls: Optional[Type[GatewayRatelimiterProto]] = None,
        shard_cls: Optional[Type[ShardProto]] = None,
        shard_ids: Optional[list[int]] = None,
        shard_count: Optional[int] = None,
    ) -> None:
        self.shards: dict[int, ShardProto] = {}
        self._shard_tasks: dict[int, Task[None]] = {}

        self._shard_ids = shard_ids
        self._shard_count = shard_count

        if shard_ids and not shard_count:
            raise ValueError("shard_count must be set if shard_ids is set")

        self._hooks: dict[str, list[Callable[..., Coroutine[Any, Any, None]]]] = {}

        self._token = token
        self._intents = intents
        self._http = http

        self._ratelimiter_cls = ratelimiter_cls or GatewayRatelimiter
        self._shard_cls = shard_cls or Shard

        self._done = Future[None]()

    async def _get_gateway(self) -> GetGatewayBot:
        route = Route("GET", "/gateway/bot")
        response = await self._http.request(route)

        data = await response.json()

        return data

    def add_dispatch_hook(self, event: str, hook: Callable[..., Coroutine[Any, Any, None]]) -> None:
        if event not in self._hooks:
            self._hooks[event] = [hook]
            return
        self._hooks[event].append(hook)

    def get_shard(self, id: int) -> ShardProto:
        return self.shards[id]

    async def start(self, *, fail_early: bool = False, wait: bool = True) -> None:
        gateway = await self._get_gateway()
        session_limits = gateway["session_start_limit"]

        if not self._shard_ids:
            if self._shard_count:
                self._shard_ids = list(range(self._shard_count))
            else:
                self._shard_count = gateway["shards"]
                self._shard_ids = list(range(self._shard_count))

        if fail_early and session_limits["remaining"] < len(self._shard_ids):
            raise RuntimeError("Insufficient remaining identify calls to start shards.")

        if not self._shard_count:
            raise RuntimeError("shard_count is not set while starting shards")

        limiter = self._ratelimiter_cls(session_limits["max_concurrency"], 5)
        last_shard: Optional[ShardProto] = None

        for shard_id in self._shard_ids:
            shard = self._shard_cls(
                shard_id,
                self._shard_count,
                self._token,
                self._intents,
                self._ratelimiter_cls,
            )

            shard.callbacks.update(self._hooks)
            self.shards[shard_id] = shard

            await limiter.acquire()

            if last_shard:
                await last_shard.ready.wait()

            last_shard = shard

            self._shard_tasks[shard.id] = create_task(shard.connect(gateway["url"], self._http))

        if wait:
            await self._done

    async def close(self) -> None:
        for shard in self.shards.values():
            await shard.close()

        self._done.set_result(None)

    async def __aenter__(self) -> "GatewayClient":
        await self.start()
        return self

    async def __aexit__(self, *exc: Any) -> None:
        await self.close()



from .client import GatewayClientProto
from .ratelimiter import GatewayRatelimiterProto
from .shard import Command, ShardProto

__all__ = (
    "Command",
    "GatewayClientProto",
    "GatewayRatelimiterProto",
    "ShardProto",
)



from asyncio import sleep
from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Final, Mapping, Optional, Type

from aiohttp import ClientResponse, ClientSession, ClientWebSocketResponse, FormData
from json import dumps

from orx.impl.errors import (
    BadGateway,
    BadRequest,
    Forbidden,
    GatewayTimeout,
    HTTPError,
    MethodNotAllowed,
    NotFound,
    OrxError,
    ServerError,
    ServiceUnavailable,
    TooManyRequests,
    Unauthorized,
    UnprocessableEntity,
)
from orx.impl.types import UNSET, UnsetOr
from orx.proto.http import RatelimiterProto, RouteProto

from .file import File
from .ratelimiter import Ratelimiter

API_URL: Final[str] = "https://discord.com/api/v9"
MAX_RETRIES: Final[int] = 3


@dataclass(slots=True)
class _RequestData:
    data: Optional[FormData]
    json: UnsetOr[Any]


class HTTPClient:
    __slots__ = (
        "_token",
        "_api_url",
        "_default_headers",
        "_max_retries",
        "_ratelimiter",
        "__session",
    )

    _status_codes: Mapping[int, Type[HTTPError]] = defaultdict(
        lambda: HTTPError,
        {
            400: BadRequest,
            401: Unauthorized,
            403: Forbidden,
            404: NotFound,
            405: MethodNotAllowed,
            422: UnprocessableEntity,
            429: TooManyRequests,
            500: ServerError,
            502: BadGateway,
            503: ServiceUnavailable,
            504: GatewayTimeout,
        },
    )

    def __init__(
        self,
        token: Optional[str] = None,
        *,
        api_url: Optional[str] = None,
        default_headers: Optional[dict[str, str]] = None,
        max_retries: Optional[int] = None,
        ratelimiter: Optional[RatelimiterProto] = None,
    ) -> None:
        self._token = token
        self._api_url = api_url or API_URL
        self._default_headers = default_headers or {}
        self._max_retries = max_retries or 3
        self._ratelimiter = ratelimiter or Ratelimiter()

        self.__session: Optional[ClientSession] = None

    async def __aenter__(self) -> "HTTPClient":
        return self

    async def __aexit__(self, *exc: Any) -> None:
        await self.close()

    @property
    def _session(self) -> ClientSession:
        if self.__session is None or self.__session.closed:
            self.__session = ClientSession(headers=self._default_headers)

        return self.__session

    @staticmethod
    def _perpare_data(attempt: int, files: UnsetOr[list[File]] = UNSET, json: UnsetOr[Any] = UNSET) -> _RequestData:
        if not files:
            return _RequestData(None, json)

        data = FormData()

        for i, file in enumerate(files):
            file.reset(attempt)

            data.add_field(f"file_{i}", file.fp, filename=file.filename)

        if json is not UNSET:
            data.add_field("payload_json", dumps(json), content_type="application/json")

        return _RequestData(data, UNSET)

    async def close(self) -> None:
        if self.__session and not self.__session.closed:
            await self.__session.close()

    async def spawn_websocket(self, url: str, **kwargs: Any) -> ClientWebSocketResponse:
        return await self._session.ws_connect(url, **kwargs)

    async def request(
        self,
        route: RouteProto,
        query_params: Optional[dict[str, str | int]] = None,
        headers: Optional[dict[str, str]] = None,
        *,
        max_retries: Optional[int] = None,
        files: UnsetOr[list[File]] = UNSET,
        json: UnsetOr[Any] = UNSET,
        reason: Optional[str] = None,
    ) -> ClientResponse:
        query_params = query_params or {}
        headers = headers or {}
        max_retries = max_retries or self._max_retries

        if self._token:
            headers["Authorization"] = f"Bot {self._token}"

        if reason:
            headers["X-Audit-Log-Reason"] = reason

        for attempt in range(max_retries or 1):
            await sleep(attempt ** 2 * 0.5)

            data = self._perpare_data(attempt, files, json)

            kwargs: dict[str, Any] = {}

            if data.data:
                kwargs["data"] = data.data
            elif data.json is not UNSET:
                kwargs["json"] = data.json

            async with await self._ratelimiter.acquire(route.bucket) as bucket:
                response = await self._session.request(
                    route.method,
                    f"{self._api_url}{route.url}",
                    params=query_params,
                    headers=headers,
                    **kwargs,
                )

                rl_reset_after = float(response.headers.get("X-RateLimit-Reset-After", 0))
                rl_bucket_remaining = int(response.headers.get("X-RateLimit-Remaining", 1))

                if 200 <= response.status < 300:
                    if rl_bucket_remaining == 0:
                        await bucket.defer(rl_reset_after)
                    return response

                if response.status == 429:
                    if "Via" not in response.headers:
                        # When a request goes through Discord's servers, and thus Google's
                        # load balancers, those load balancers add a `Via` header to the response.
                        # If we are ratelimited by Cloudflare then this header won't have been
                        # added by GCP, so we raise.

                        raise TooManyRequests(response)

                    response_json: dict[str, Any] = await response.json()

                    is_global = response_json.get("global", False)
                    retry_after = response_json["retry_after"]

                    if is_global:
                        await self._ratelimiter.set_global_lock(retry_after)
                    else:
                        await bucket.defer(retry_after)
                elif response.status < 500:
                    raise self._status_codes[response.status](response)

        raise OrxError(f"Failed to make request on route {route.method} {route.url} after {max_retries} attempts.")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['orx',
 'orx.impl',
 'orx.impl.gateway',
 'orx.impl.http',
 'orx.proto',
 'orx.proto.gateway',
 'orx.proto.http']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp[speedups]>=3.8.1,<4.0.0', 'discord-typings>=0.4.0,<0.5.0']

setup_kwargs = {
    'name': 'orx',
    'version': '1.0.0a0',
    'description': 'A modern, async, lightweight Discord API wrapper.',
    'long_description': "# orx\n\nA modern, async, lightweight Discord API wrapper.\n\n## Key Features\n\n- Efficient and robust ratelimit handling\n- Consistently and accurately typed\n- Provides low-level access for customisability\n\n### What Orx is **not**\n\n- A high level wrapper for people who just want a simple API wrapper, for that use discord.py or a similar fork of it.\n\n## Installing\n\n**Orx requires Python 3.10 or higher.**\n\nTo install Orx you can use:\n\n```sh\n# Linux/MacOS\npython3 -m pip install orx\n\n# Windows\npy -3 -m pip install Orx\n```\n\n## Versioning\n\nOrx is versioned according to semantic versioning with minor modifications.\n\nThe version format is `major.minor.patch`, i.e. `1.0.0`.\n\n- Major version changes may have large breaking changes to how the library works.\n- Minor version changes may have small breaking changes, for example a single method's signature changing.\n- Patch version changes should not include breaking changes.\n\nSome versions released on pypi may use the a format similar to `1.0.0a0` or `1.0.0rc0`. These are alpha and release candidate versions. It is not recommended that you use these versions in production instances of your bots.\n\n## Contributing\n\nSee [Contributing to Orx](./CONTRIBUTING.md).\n",
    'author': 'vcokltfre',
    'author_email': 'vcokltfre@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/vcokltfre/orx',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from asyncio import Event, Task, create_task, sleep
from random import randrange
from time import time
from typing import Any, Callable, Coroutine, Optional, Type, cast

from aiohttp import ClientWebSocketResponse, WSMsgType
from discord_typings.gateway import GatewayEvent as GatewayEventType
from discord_typings.gateway import (
    HeartbeatCommand,
    HelloEvent,
    IdentifyCommand,
    IdentifyConnectionProperties,
    IdentifyData,
    ResumeCommand,
    ResumeData,
)

from orx.impl.errors import GatewayCriticalError, GatewayReconnect
from orx.proto.gateway import Command, GatewayRatelimiterProto
from orx.proto.http import HTTPClientProto

from .enums import GatewayCloseCodes, GatewayOps
from .event import GatewayEvent

CRITICAL = [
    GatewayCloseCodes.NOT_AUTHENTICATED,
    GatewayCloseCodes.AUTHENTICATION_FAILED,
    GatewayCloseCodes.INVALID_API_VERSION,
    GatewayCloseCodes.INVALID_INTENTS,
    GatewayCloseCodes.DISALLOWED_INTENTS,
]

NONCRITICAL = [
    GatewayCloseCodes.INVALID_SEQ,
    GatewayCloseCodes.RATE_LIMITED,
    GatewayCloseCodes.SESSION_TIMEOUT,
]


class Shard:
    __slots__ = (
        "callbacks",
        "id",
        "latency",
        "ready",
        "_count",
        "_http",
        "_token",
        "_intents",
        "_ratelimiter",
        "_session",
        "_sequence",
        "_socket",
        "_pacemaker",
        "_last_hb",
        "_last_ack",
        "_closing",
    )

    def __init__(
        self,
        id: int,
        shard_count: int,
        token: str,
        intents: int,
        ratelimiter_cls: Type[GatewayRatelimiterProto],
    ) -> None:
        self.id = id
        self.callbacks: dict[str, list[Callable[..., Coroutine[Any, Any, None]]]] = {}

        self.ready = Event()
        self.ready.clear()

        self._count = shard_count
        self._token = token
        self._intents = intents

        self._ratelimiter = ratelimiter_cls(120, 60)

        self._session: Optional[str] = None
        self._sequence: Optional[int] = None

        self._socket: Optional[ClientWebSocketResponse] = None

        self._pacemaker: Optional[Task[None]] = None
        self._last_hb: Optional[float] = None
        self._last_ack: Optional[float] = None

        self._closing: bool = False

        self.latency: Optional[float] = None

    def __repr__(self) -> str:
        return f"<Shard id={self.id} seq={self._sequence}>"

    async def _resume(self) -> None:
        if not (self._session and self._sequence):
            raise RuntimeError("Cannot resume shard with no session or sequence.")

        await self.send(
            ResumeCommand(
                op=GatewayOps.RESUME,  # type: ignore
                d=ResumeData(
                    token=self._token,
                    session_id=self._session,
                    seq=self._sequence,
                ),
            )
        )

    async def _identify(self) -> None:
        await self.send(
            IdentifyCommand(
                op=GatewayOps.IDENTIFY,  # type: ignore
                d=IdentifyData(
                    token=self._token,
                    intents=self._intents,
                    shard=[self.id, self._count],
                    properties=IdentifyConnectionProperties(
                        **{
                            "$os": "linux",
                            "$browser": "orx",
                            "$device": "orx",
                        }
                    ),
                ),  # type: ignore
            )
        )

    def _task_error(self, task: Task[None]) -> None:
        exc = task.exception()

        if not exc:
            return

        raise exc

    def _callback(self, event: GatewayEvent) -> None:
        for callback in self.callbacks.get(event.dispatch_name, []):
            task = create_task(callback(event))
            task.add_done_callback(self._task_error)

        for callback in self.callbacks.get("*", []):
            task = create_task(callback(event))
            task.add_done_callback(self._task_error)

    async def _dispatch(self, data: GatewayEventType) -> None:
        event = GatewayEvent(self, data["op"], data, data.get("s"), data.get("t"))

        try:
            self._callback(event)
        except Exception as e:
            print(f"Error in callback: {e}")

        if event.op == GatewayOps.HELLO:
            data = cast(HelloEvent, event.data)
            self._pacemaker = create_task(self._heartbeat(data["d"]["heartbeat_interval"]))
            self._pacemaker.add_done_callback(self._task_error)
            await self._identify()
        elif event.op == GatewayOps.ACK:
            self._last_ack = time()
        elif event.op == GatewayOps.RECONNECT:
            await self.close()
            raise GatewayReconnect()
        elif event.t == "READY":
            self.ready.set()

    async def _handle_disconnect(self, code: int) -> None:
        if code in CRITICAL:
            raise GatewayCriticalError(code)

        if code in NONCRITICAL:
            self._session = None
            self._sequence = None

        await self.close()

        raise GatewayReconnect()

    async def _heartbeat(self, delay: float) -> None:
        await sleep(randrange(0, int(delay)))

        while not self._closing:
            if self._last_ack and time() - self._last_ack >= delay:
                return await self.close()

            await self.send(HeartbeatCommand(op=1, d=self._sequence))
            self._last_hb = time()

            await sleep(delay)

    async def _read(self) -> None:
        if not self._socket or self._socket.closed:
            raise RuntimeError("Shard is not connected")

        async for message in self._socket:
            if message.type == WSMsgType.TEXT:  # type: ignore
                data: GatewayEventType = message.json()

                if sequence := data.get("s"):
                    if (not self._sequence) or self._sequence < sequence:
                        self._sequence = sequence

                await self._dispatch(data)

        await self._handle_disconnect(self._socket.close_code or 1000)

    async def connect(self, url: str, http: HTTPClientProto) -> None:
        if self._socket and not self._socket.closed:
            raise RuntimeError("Shard is already connected")

        self._socket = await http.spawn_websocket(url + "?v=10")

        if self._session:
            await self._resume()

        try:
            await self._read()
        except GatewayReconnect:
            pass

        if self._closing:
            return

    async def close(self) -> None:
        self._closing = True

        if self._socket and not self._socket.closed:
            await self._socket.close()

        if self._pacemaker and not self._pacemaker.cancelled():
            self._pacemaker.cancel()

    async def send(self, data: Command) -> None:
        if not self._socket or self._socket.closed:
            raise RuntimeError("Shard is not connected")

        await self._ratelimiter.acquire()

        try:
            await self._socket.send_json(data)
        except OSError:
            await self.close()
        except Exception:
            await self.close()
            raise



from typing import Any, Callable, Coroutine, Optional, Protocol, Type

from orx.proto.http import HTTPClientProto

from .ratelimiter import GatewayRatelimiterProto
from .shard import ShardProto


class GatewayClientProto(Protocol):
    shards: dict[int, ShardProto]

    def __init__(
        self,
        token: str,
        intents: int,
        http: HTTPClientProto,
        ratelimiter_cls: Optional[Type[GatewayRatelimiterProto]] = None,
        shard_cls: Optional[Type[ShardProto]] = None,
        shard_ids: Optional[list[int]] = None,
        shard_count: Optional[int] = None,
    ) -> None:
        ...

    def add_dispatch_hook(self, event: str, hook: Callable[..., Coroutine[Any, Any, None]]) -> None:
        ...

    def get_shard(self, id: int) -> ShardProto:
        ...

    async def start(self) -> None:
        ...

    async def close(self) -> None:
        ...
